#!perl

use threads;
use threads::shared;
use Date::Parse;
use Date::Format;
use NNML::Server;
use NNML::Config qw($Config);
use NNML::Active qw($ACTIVE);

use lib::Config;
use lib::NNmirror;
use lib::TimeZone;
use lib::NNTP;
use lib::Control;

require $config;
require $hooks;

my $flag : shared = 1;
my $inloop : shared = 0;
my $loops = 2;
my $interval = shift @ARGV;
my $prefixed = "host|port|user|pass";
my $optexp = "group|" . $prefixed;
my @admin, @feed;

open(USERS, "<$users") or die "Cannot open $users: $!";
while (<USERS>) {
    next if /^$/ or /^#/;
    @admin = (split)[0, 1] if /\sadmin/;
    @feed = (split)[0, 1] if /^feed/;
}
close(USERS);

my $server = threads->new(\&server);
my $feed = threads->new(\&feed_loop, $interval);
my $daily_job = threads->new(\&daily);

$_ = <> if $interval;

sleep 1;
$flag = 0 if $inloop >= $loops;

end_proc($server, $feed, $daily_job);

# end of main


# thread for feed
sub feed_loop {
    my $interval = shift;
    my ($group, $upd) = fetch_ngs();

    update_ngs($Config->{ngs}, $group, $upd);
    update_active($group);

    print "feed loop\n";

    for my $s (keys %site) {
	push(@ropt, nnopts($s, 'f')) if $site{$s}->{mode} =~ /r/;
	push(@wopt, nnopts($s, 't')) if $site{$s}->{mode} =~ /w/;
    }

    while ($flag) {
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)
	    = localtime(time - 3600);
	my $update = sprintf("%04d%02d%02d", 1900 + $year, ++$mon, $mday);
	my $uptime = sprintf("%02d%02d%02d", $hour, $min, $sec);

	++$inloop if $inloop < $loops;

	call_nnmirror($update, $uptime, @ropt);
	print STDERR "NNML::Server::unspool\n";
	unspool;
	call_nnmirror($update, $uptime, @wopt);

	if ($interval and $flag) {
	    print STDERR "waiting $interval seconds\n";
	    for (1..$interval) {
		sleep 1;
		last if ! $flag;
	    }
	} else {
	    $flag = 0;
	}
    }
}

sub update_active {
    my $ngs = shift;

    print "update active\n";

    my %group = (qw[local.test] => "y");
    for my $s (keys %site) {
	next if $site{$s}->{mode} !~ /r/;

	my $nntp = newNNTP($site{$s}->{host}, $site{$s});
	my @grouppat = split /,/, $site{$s}->{group};
	my $pos = join ',', grep /^[^!]/, @grouppat;
	my $neg = join '|', grep s/^!//, @grouppat;
	$neg =~ s/\./\\./g;
	$neg =~ s/\*/.*/g;
	for (split ',', $pos) {
	    my $g = $nntp->active($_);
	    for (keys %{$g}) {
		next if $neg and /^($neg)$/;
		$group{$_} = $g->{$_}->[2];
	    }
	}
	$nntp->quit;
    }
    for (sort keys %group) {
	next if $ACTIVE->group($_);
#	print join(": ", ($_, $group{$_}, $ngs->{$_})) . "\n";
	$ACTIVE->new_group($_, $group{$_}, $ngs->{$_});
    }
}

sub update_ngs {
    my ($ngs, $group, $upd) = @_;

    return unless $upd;

    print "update newsgroups\n";

    open(NGS, ">$ngs") or die "Cannot open $ngs: $!";
    for (sort keys %{$group}) {
	print NGS "$_\t" . $group->{$_} . "\n";
    }
    close(NGS);

    return;
}

sub fetch_ngs {
    my $group = {qw[local.test] => "test in localhost."};
    my $upd = 0;

    if (-f $ngs) {
	open(NGS, "<$ngs") or die "Cannot open $ngs: $!";
	while (<NGS>) {
	    chomp;
	    my ($gr, $ds) = split /\t+/;
	    $group->{$gr} = $ds;
	}
	close(NGS);
    }

    for my $s (keys %site) {
	next if $site{$s}->{mode} !~ /r/;

	my $nntp = newNNTP($site{$s}->{host}, $site{$s});
	my @grouppat = split /,/, $site{$s}->{group};
	my $pos = join ',', grep /^[^!]/, @grouppat;
	my $neg = join '|', grep s/^!//, @grouppat;
	$neg =~ s/\./\\./g;
	$neg =~ s/\*/.*/g;
	for (split ',', $pos) {
	    my $g = $nntp->newsgroups($_);
	    for (keys %{$g}) {
		next if $neg and /^($neg)$/;
		next if $group->{$_} eq $g->{$_};
		$group->{$_} = $g->{$_};
		$upd = 1;
	    }
	}
	$nntp->quit;
    }

    return ($group, $upd);
}

sub call_nnmirror {
    my $update = shift;
    my $uptime = shift;
    my %date, %time;

    open(DATE, "<$datefile");
    while (<DATE>) {
	my ($fromto, $date, $time) = split;
	$date{$fromto} = $date;
	$time{$fromto} = $time;
    }
    close(DATE);
    for my $opt (@_) {
	my $fromto = ${$opt}{fhost} . ":" . ${$opt}{thost};
	$date{$fromto} = "19700101" if ! $date{$fromto};
	$time{$fromto} = "090000" if ! $time{$fromto};
	${$opt}{date} = $date{$fromto};
	${$opt}{time} = $time{$fromto};
	my @opts;
	for $o (keys %{$opt}) {
	    push @opts, "$o => " . ${$opt}{$o};
	}
	print STDERR "nnmirror(" . join(", ", @opts) . ")\n";
	next if nnmirror(%{$opt});
	$date{$fromto} = $update;
	$time{$fromto} = $uptime;
    }
    open(DATE, ">$datefile");
    for (sort keys %date) {
	print DATE join(" ", ($_, $date{$_}, $time{$_})) . "\n";
    }
    close(DATE);
}

sub nnopts {
    my ($s, $ft) = @_;
    my $tf = ($ft eq 't') ? 'f' : 't';
    my %opt;

    $opt{"${tf}host"} = "localhost";
    $opt{"${tf}user"} = $feed[0];
    $opt{"${tf}pass"} = $feed[1];

    for (keys %{$site{$s}}) {
	next if ! /$optexp/;
	$val = ${$site{$s}}{$_};
	s/($prefixed)/$ft$1/;
	$opt{$_} = $val;
    }

    return \%opt;
}

# thread for daily job
sub daily {
    print "daily job\n";

    open(DAILY, "<$daily");
    my $d = <DAILY>;
    close(DAILY);
    chomp $d;
    my $next_time = str2time($d) + 24 * 3600;

    while ($flag) {
	++$inloop if $inloop < $loops;
	while ($flag) {
	    last if time > $next_time;
	    for (1..3600) {
		last if ! $flag;
		sleep 1;
	    }
	}
	last if ! $flag;

	my $ct = ctime(time);
	chomp $ct;
	open(DAILY, ">$daily");
	print DAILY "$ct $TZ\n";
	close(DAILY);
	$next_time = time + 24 * 3600;

#	update_ngs($Config->{ngs}, fetch_ngs());
	expire();
    }
}

sub expire {
    my $nntp = newNNTP('localhost', user => $admin[0], pass => $admin[1]);
    my $exgroup = '';
    my @exipre;
    my $gr, $ex;

    open(EX, "<$expiretxt");
    while (<EX>) {
	next if /^$/ or /^#/;
	chomp;
	push @expire, $_;
    }
    close(EX);

    $exgroup = '';
    open (EXL, ">$expirelist");
    while ($_ = pop @expire) {
	($gr, $ex) = split /:/;
	next if $ex eq 'never' or $ex < 1;

	my $extime = time - $ex * 24 * 3600; # $ex days ago
	my $group = $nntp->newsgroups($gr . $exgroup);

	for my $g (keys %{$group}) {
	    my @mygroup = $nntp->group($g);
	    my $xover = $nntp->xover(($mygroup[1], $mygroup[2]));
	    for (sort keys %$xover) {
		my $atime = str2time($xover->{$_}->[2]);
		next if $atime > $extime or ! $atime;
		print EXL "$g:$_: " . $xover->{$_}->[2] . ": " .
		    str2time($xover->{$_}->[2]) . "\n";
		$nntp->command('XDELETE', $_)->response;
	    }
	}
    } continue {
	$exgroup .= ",!$gr";
    }
    close(EXL);
    $nntp->quit;
}

sub end_proc {
    my ($server, $feed, $daily_job) = @_;

    $feed->join;
    $daily_job->join;

    my $nntp = newNNTP('localhost', user => $admin[0], pass => $admin[1]);
    $nntp->command('SHUT');
    $nntp->quit;
    $server->join;
    print STDERR "server finished\n"
}

# end of file
